/* This file is part of XTee Keyboard, a tool by Alessandro Pira */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define HEIGHT 200
#define WIDTH 200

//#define FIXED_SIZE

#define APPNAME "xteekeyb"
#define APPTITLE "XTee Keyboard"

static Window *wtgt;
static int wtgt_size;
static Display *xdpy;
static int (*old_err_handler)(Display *, XErrorEvent *);

int isnum(char *str)
{
	while (*str != '\0')
	{
		if (!isdigit(*(str++)))
		{
			return 0;
		}
	}
	return 1;
}

void help(char *argv0)
{
	printf("\n  " APPTITLE "\n\nUsage:\n\n");
	printf("%s -h\n", argv0);
	printf("    prints this help\n");
	printf("%s search [WORD [WORD...]]\n", argv0);
	printf("    searches for window IDs by window title\n");
	printf("%s [WINDOW_ID [WINDOW_ID...]]\n", argv0);
	printf("    starts the main app forwarding keystrokes to all the listed windows\n\n");
}

void search_win(char *title[], int titlecnt, Window parent, int indent)
{
	int i, j, c, rv, idx=0;

	char **list = NULL;
	XTextProperty tp;

	rv = 0;

	XGetWMName(xdpy, parent, &tp);
	XmbTextPropertyToTextList(xdpy, &tp, &list, &c);

	int found = 0;
	if (titlecnt <= 0)
	{
		idx = (c > 0) ? 0 : -1;
		found = 1;
	}
	else
	{
		for (i = 0; i < c; i++)
		{
			for (j = 0; j<titlecnt; j++)
			{
				if (strstr(list[i], title[j]) != NULL)
				{
					found = 1;
					idx = i;
					break;
				}
			}
			if (found) break;
		}
	}

	if (found)
	{
		int x;
		for (x=0; x<indent; x++) printf("  ");
		if (idx >= 0)
		{
			printf("%lu %s\n", parent, list[idx]);
		}
		else
		{
			printf("%lu [UNKNOWN]\n", parent);
		}
		indent++;
	}

	XFreeStringList(list);
	XFree(tp.value);

	Window dummy;
	Window *child;
	unsigned int nchild;

	if (!XQueryTree(xdpy, parent, &dummy, &dummy, &child, &nchild))
	{
		return;
	}

	for (i = 0; i < nchild; i++)
	{
		search_win(title, titlecnt, child[i], indent);
	}

	XFree(child);
	return;
}


int err_handler(Display *xdpy, XErrorEvent *event)
{
	if (event->error_code != BadWindow)
	{
		return (*old_err_handler)(xdpy, event);
	}
	return 0;
}

void do_send(int idx, XEvent *event, long mask)
{
	if (wtgt[idx] != 0)
	{
		((XKeyEvent*)event)->window = wtgt[idx];
		XSendEvent(xdpy, wtgt[idx], 1, mask, event);
	}
}

void xtee_run()
{
	int i;

	xdpy = XOpenDisplay(NULL);

	Window w;
	XColor bg_col;
	Colormap colormap;
	char bg[] = "#007DAC";
	XClassHint *ch;
#ifdef FIXED_SIZE
	XSizeHints *sh;
#endif

	colormap = DefaultColormap(xdpy, 0);

	XParseColor(xdpy, colormap, bg, &bg_col);
	XAllocColor(xdpy, colormap, &bg_col);

	w = XCreateSimpleWindow(xdpy, RootWindow(xdpy, 0),
		40, 40, WIDTH, HEIGHT, 2, BlackPixel(xdpy, 0),
		bg_col.pixel);

	ch = XAllocClassHint();
	ch->res_name = APPNAME;
	ch->res_class = APPNAME ".MainWindow";
	XSetClassHint(xdpy, w, ch);

#ifdef FIXED_SIZE
	sh = XAllocSizeHints();
	sh->width = WIDTH;
	sh->height = HEIGHT;
	sh->min_width = WIDTH;
	sh->min_height = HEIGHT;
	sh->max_width = WIDTH;
	sh->max_height = HEIGHT;
	sh->flags = PSize|PMinSize|PMaxSize;
	XSetWMSizeHints(xdpy, w, sh, XInternAtom(xdpy, "WM_NORMAL_HINTS", 0));
#endif

	Atom wmDelete=XInternAtom(xdpy, "WM_DELETE_WINDOW", True);
	XSetWMProtocols(xdpy, w, &wmDelete, 1);

/*	GC gc;
	gc = XCreateGC(xdpy, w, 0, 0);
	XSetForeground(xdpy, gc, bg_col.pixel);
*/

	XStoreName(xdpy, w, APPTITLE);
	XSelectInput(xdpy, w, KeyPressMask | KeyReleaseMask | ExposureMask | DestroyNotify);

	XMapWindow(xdpy, w);

	old_err_handler = XSetErrorHandler(&err_handler);

	XFlush(xdpy);

	int done = 0;

	for (done = 0; !done; ) {
		XEvent event;
		XNextEvent (xdpy, &event);

		switch (event.type)
		{
		case KeyPress:
//			printf("P: %d\n", event.xkey.keycode);
			for (i=0; i<wtgt_size; i++)
			{
				do_send(i, &event, KeyPressMask);
			}
			break;
		case KeyRelease:
//			printf("R: %d\n", event.xkey.keycode);
			for (i=0; i<wtgt_size; i++)
			{
				do_send(i, &event, KeyReleaseMask);
			}
			break;
		case Expose:
//			XFillRectangle(xdpy, w, gc, 0, 0, 200, 200);
			break;
		case DestroyNotify:
		case ClientMessage:
			done = 1;
			break;
		default:
			break;
		}

		XFlush(xdpy);
	}

#ifdef FIXED_SIZE
	XFree(sh);
#endif
	XFree(ch);
	XUnmapWindow(xdpy, w);
	XDestroyWindow(xdpy, w);
	XCloseDisplay(xdpy);
}

int main(int argc, char *argv[])
{
	if ((argc <= 1) || (strcmp("-h", argv[1]) == 0))
	{
		help(argv[0]);
	}
	else if (strcmp("search", argv[1]) == 0)
	{
		xdpy = XOpenDisplay(NULL);
		search_win(&(argv[2]), argc - 2, XDefaultRootWindow(xdpy), 0);
		XCloseDisplay(xdpy);
	}
	else
	{
		int ok = 1;
		int i;

		for (i=1; i<argc; i++)
		{
			if (!isnum(argv[i]))
			{
				ok = 0;
				break;
			}
		}

		if (ok)
		{
			wtgt_size = argc - 1;
			wtgt = malloc(sizeof(Window) * wtgt_size);

			for (i=1; i<argc; i++)
			{
				wtgt[i-1] = strtoul(argv[i], NULL, 0);
			}

			xtee_run();

			free(wtgt);
		}
		else
		{
			help(argv[0]);
		}
	}

	return 0;
}


